/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/


#import "BDRuleModel.h"
#import "BDRule.h"
#import "BDAssignment.h"


#pragma mark NSPredicate&NSCompoundPredicate(BDRuleModelSupport)

@interface NSPredicate(BDRuleModelSupport)
- (NSUInteger) predicateKeyCount;
@end

@implementation NSPredicate(BDRuleModelSupport)
- (NSUInteger) predicateKeyCount
{
    return 1;
}
@end

@implementation NSCompoundPredicate(BDRuleModelSupport)
- (NSUInteger) predicateKeyCount
{
    NSUInteger theResult = 0;
    
    for(NSPredicate* i in self.subpredicates)
    {
        theResult += [i predicateKeyCount];
    }
    
    return theResult;
}
@end

#pragma mark -

@implementation BDRuleModel

#pragma mark -
#pragma mark class methods

#pragma mark -
#pragma mark init methods

- (id)init
{
    return [self initWithRules:[NSMutableArray array]];
}

- (id)initWithRules:(NSArray *)inRules
{
    if (self = [super init])
    {
        // Set the rules in this model
        self.rules = [inRules mutableCopy];

        // The candidate rules for a qualifier have the requested key path as their right-hand side key path.
        candidateQualifier = [NSPredicate predicateWithFormat:@"rhs.keyPath == $keyPath"];
        
        
        // Candidate rules are sorted according to priority, from highest to lowest.
        // Within that, they are sorted by the number of qualifier keys in the left-hand
        // side.  This guarantees that the rule at the front of the returned array
        // (index 0) will be the most specific rule possible.
        NSSortDescriptor *priorityOrdering;
        NSSortDescriptor *qualifierKeysCountOrdering;

        priorityOrdering = [[NSSortDescriptor alloc] initWithKey:@"priority"
                                                        ascending:NO];
        
        qualifierKeysCountOrdering = [[NSSortDescriptor alloc] initWithKey:@"lhs.predicateKeyCount"
                                                                  ascending:NO];
        
        candidateSort = [[NSArray alloc] initWithObjects:priorityOrdering, qualifierKeysCountOrdering, nil];

    }

    return self;
}

- (id)initWithContentsOfFile:(NSString *)path
{
    id s;
    NSDictionary *plist = nil;

    plist = [NSDictionary dictionaryWithContentsOfFile:path];

    s =  [self initWithPropertyList:plist
                              owner:nil];

    [s awakeWithPropertyList:plist];

    return s;
}

#pragma -
#pragma mark dealloc

- (void)dealloc
{
    self.rules = nil;
    
    candidateQualifier = nil;
    candidateSort = nil;
}

#pragma mark -
#pragma mark NSCopying, hash, isEqual:

#pragma mark -
#pragma mark NSCoding methods
- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeObject:self.rules];
}

- (id)initWithCoder:(NSCoder *)coder
{
    NSArray *theRules = [coder decodeObject];
    
    return [self initWithRules:theRules];
}

#pragma mark -
#pragma mark BDPropertyListEncoding methods
- (id)initWithPropertyList:(NSDictionary *)propertyList
                     owner:(id)owner
{
    int version;
    NSMutableArray *theRules = nil;
    
    version = [[propertyList valueForKey:@"version"] intValue];
    
    theRules = [NSMutableArray arrayWithCapacity:0];
    
    if (version >= 1) {
        NSArray *plistRules;
        NSDictionary *onePlistRule;
        
        plistRules = [propertyList valueForKey:@"rules"];
        
        for (onePlistRule in plistRules) {
            BDRule *theRule;
            
            theRule = [[BDRule alloc] initWithPropertyList:onePlistRule
                                                     owner:self];
            
            [theRules addObject:theRule];
        }
    }
    
    return [self initWithRules:theRules];
}

- (void)awakeWithPropertyList:(NSDictionary *)propertyList
{
    int version;
    
    version = [[propertyList valueForKey:@"version"] intValue];
    
    if (version >= 1) {
        NSArray *plistRules;
        unsigned i;
        
        plistRules = [propertyList valueForKey:@"rules"];
        
        for (i = 0; i < [plistRules count]; i++)
        {
            NSDictionary *onePlistRule;
            BDRule *theRule;
            
            onePlistRule = [plistRules objectAtIndex:i];
            theRule = [self.rules objectAtIndex:i];
            
            [theRule awakeWithPropertyList:onePlistRule];
        }
    }
}

- (void)encodeIntoPropertyList:(NSMutableDictionary *)propertyList
{
    NSMutableArray *plistRules;
    NSEnumerator *rulesEnumerator;
    BDRule *oneRule;
    
    plistRules = [NSMutableArray arrayWithCapacity:0];
    
    [propertyList setValue:[NSNumber numberWithInt:1]
                     forKey:@"version"];
    
    rulesEnumerator = [self.rules objectEnumerator];
    
    while (oneRule = [rulesEnumerator nextObject]) {
        NSMutableDictionary *plistRule;
        
        plistRule = [NSMutableDictionary dictionaryWithCapacity:0];
        
        [oneRule encodeIntoPropertyList:plistRule];
        
        [plistRules addObject:plistRule];
    }
    
    [propertyList setValue:plistRules
                     forKey:@"rules"];
}

#pragma mark -
#pragma mark awakeFromNib-like methods

#pragma mark -
#pragma mark action methods

#pragma mark -
#pragma mark delegate/datasource methods

#pragma mark -
#pragma mark accessor methods (in pairs)

- (void)addToRules:(BDRule *)value
{
    [rules addObject:value];
}

- (void)removeFromRules:(BDRule *)value
{
    [rules removeObject:value];
}

@synthesize rules;

#pragma mark -
#pragma mark Utility methods

- (NSArray *)candidateRulesForKey:(NSString *)key
{
    NSArray *candidateRules = nil;
    NSPredicate *qualifier;
    NSArray *matchingRules;
    
    qualifier = [candidateQualifier predicateWithSubstitutionVariables:
                 [NSDictionary dictionaryWithObject:key
                                             forKey:@"keyPath"]];
    
    matchingRules = [self.rules filteredArrayUsingPredicate:qualifier];
    
    candidateRules = [matchingRules sortedArrayUsingDescriptors:candidateSort];
    
    return candidateRules;
}


- (BOOL)writeToFile:(NSString *)path
         atomically:(BOOL)flag
{
    BOOL wroteFile = NO;
    NSMutableDictionary *plist = nil;

    plist = [NSMutableDictionary dictionaryWithCapacity:0];

    [self encodeIntoPropertyList:plist];

    wroteFile = [plist writeToFile:path
                        atomically:flag];

    return wroteFile;
}


@end
