/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/

#import <BDRuleEngine/BDRuleEngine.h>

#import "TestBDRuleEngine.h"
#import "Stock.h"


@implementation TestBDRuleEngine

- (void)setUp
{
    [super setUp];

    BDRule *upRule, *downRule, *defaultColorRule;
    BDRule *redHueRule, *greenHueRule, *defaultHueRule;
    
    model = [[BDRuleModel alloc] init];

    // Add a bunch of rules to the model here.

    // (stock.differenceFromPreviousDayClose > 0) => color = "green" [0]
    upRule = [BDRule ruleWithLhs:[NSPredicate predicateWithFormat:@"stock.differenceFromPreviousDayClose > 0"]
                             rhs:[BDAssignment assignmentWithKeyPath:@"color"
                                                               value:@"green"]
                        priority:0];
    
    // (stock.differenceFromPreviousDayClose < 0) => color = "red" [0]
    downRule = [BDRule ruleWithLhs:[NSPredicate predicateWithFormat:@"stock.differenceFromPreviousDayClose < 0"]
                               rhs:[BDAssignment assignmentWithKeyPath:@"color"
                                                                 value:@"red"]
                          priority:0];

    // *true* => color = "black" [0]
    defaultColorRule = [BDRule ruleWithLhs:[NSPredicate predicateWithFormat:@"TRUEPREDICATE"]
                                       rhs:[BDAssignment assignmentWithKeyPath:@"color"
                                                                         value:@"black"]
                                  priority:0];

    // (color = 'red') => hue = 0.0
    redHueRule = [BDRule ruleWithLhs:[NSPredicate predicateWithFormat:@"color = 'red'"]
                                 rhs:[BDAssignment assignmentWithKeyPath:@"hue"
                                                                   value:[NSNumber numberWithDouble:0.0]]
                            priority:0];

    // (color = 'green') => hue = 120.0
    greenHueRule = [BDRule ruleWithLhs:[NSPredicate predicateWithFormat:@"color = 'green'"]
                                   rhs:[BDAssignment assignmentWithKeyPath:@"hue"
                                                                     value:[NSNumber numberWithDouble:120.0]]
                              priority:0];

    // *true* => hue = 240.0 [0]
    defaultHueRule = [BDRule ruleWithLhs:[NSPredicate predicateWithFormat:@"TRUEPREDICATE"]
                                     rhs:[BDAssignment assignmentWithKeyPath:@"hue"
                                                                       value:[NSNumber numberWithDouble:240.0]]
                                priority:0];
    
    [model setRules:[NSMutableArray arrayWithObjects:upRule,
                     downRule, defaultColorRule,
        redHueRule, greenHueRule, defaultHueRule, nil]];
    
    // Set up some stocks here.

    aaaa = [[Stock alloc] initWithSymbol:@"AAAA"
                        previousDayClose:14.47];

    bbbb = [[Stock alloc] initWithSymbol:@"BBBB"
                        previousDayClose:48.21];

    cccc = [[Stock alloc] initWithSymbol:@"CCCC"
                        previousDayClose:1.14];

    [aaaa setLastTrade:20.0];
    [bbbb setLastTrade:40.0];
    [cccc setLastTrade:1.14];
}

- (void)tearDown
{
    model = nil;

    aaaa = nil;
    bbbb = nil;
    cccc = nil;
    
    [super tearDown];
}

- (void)testSimpleRuleSet
{
    BDRuleContext *context;
    NSString *color;

    context = [[BDRuleContext alloc] initWithModel:model];

    [context setValue:aaaa forKey:@"stock"];
    color = [context valueForKey:@"color"];
    XCTAssertEqualObjects(color, @"green", @"");

    [context setValue:bbbb forKey:@"stock"];
    color = [context valueForKey:@"color"];
    XCTAssertEqualObjects(color, @"red", @"");

    [context setValue:cccc forKey:@"stock"];
    color = [context valueForKey:@"color"];
    XCTAssertEqualObjects(color, @"black", @"");
}

- (void)testSingleRecursiveRule
{
    BDRuleContext *context;
    NSString *hue;

    context = [[BDRuleContext alloc] initWithModel:model];

    [context setValue:aaaa forKey:@"stock"];
    hue = [context valueForKey:@"hue"];
    XCTAssertTrue([hue isEqualTo:[NSNumber numberWithDouble:120.0]], @"");

    [context setValue:bbbb forKey:@"stock"];
    hue = [context valueForKey:@"hue"];
    XCTAssertTrue([hue isEqualTo:[NSNumber numberWithDouble:0.0]], @"");

    [context setValue:cccc forKey:@"stock"];
    hue = [context valueForKey:@"hue"];
    XCTAssertTrue([hue isEqualTo:[NSNumber numberWithDouble:240.0]], @"");

}

- (void)testRuleDescriptions
{
    NSArray *rules = [model rules];

    XCTAssertEqualObjects([[rules objectAtIndex:0] description],
                         @"(stock.differenceFromPreviousDayClose > 0) => color = \"green\" [0]", @"");
    
    XCTAssertEqualObjects([[rules objectAtIndex:1] description],
                         @"(stock.differenceFromPreviousDayClose < 0) => color = \"red\" [0]", @"");
    
    XCTAssertEqualObjects([[rules objectAtIndex:2] description],
                         @"(TRUEPREDICATE) => color = \"black\" [0]", @"");
    
    XCTAssertEqualObjects([[rules objectAtIndex:3] description],
                         @"(color == \"red\") => hue = 0 [0]", @"");
    
    XCTAssertEqualObjects([[rules objectAtIndex:4] description],
                         @"(color == \"green\") => hue = 120 [0]", @"");
    
    XCTAssertEqualObjects([[rules objectAtIndex:5] description],
                         @"(TRUEPREDICATE) => hue = 240 [0]", @"");
}

- (void)testValuesForKeyPathTakingSuccessiveValuesForKeyPath
{
    BDRuleContext *context;
    NSArray *colors;

    context = [[BDRuleContext alloc] initWithModel:model];
    
    colors = [context valuesForKeyPath:@"color"
                takingSuccessiveValues:[NSArray arrayWithObjects:aaaa, bbbb, cccc, nil]
                            forKeyPath:@"stock"];

    XCTAssertEqualObjects([colors objectAtIndex:0], @"green", @"");
    XCTAssertEqualObjects([colors objectAtIndex:1], @"red", @"");
    XCTAssertEqualObjects([colors objectAtIndex:2], @"black", @"");

}

//- (void)testRuleModelFromPlistRep
//{
//    NSString *plistRep;
//    BDRuleModel *myModel;
//    BDRuleContext *context;
//
//    plistRep = @"(version = 1, (\
//                      {\
//                          lhs = \"color = 'red'\";\
//                          rhsKeyPath = saturation;\
//                          rhsValue = \"75\";\
//                          assignmentClass = BDAssignment;\
//                          priority = 0;\
//                      },\
//                      {\
//                          lhs = \"color = 'green'\";\
//                          rhsKeyPath = saturation;\
//                          rhsValue = \"50\";\
//                          assignmentClass = BDAssignment;\
//                          priority = 0;\
//                      },\
//                      {\
//                          lhs = \"color = 'blue'\";\
//                          rhsKeyPath = saturation;\
//                          rhsValue = \"66\";\
//                          assignmentClass = BDAssignment;\
//                          priority = 0;\
//                      }\
//                      ))";
// 
//    myModel = [[BDRuleModel alloc] initWithPropertyList:[plistRep propertyList] owner:self];
//
//    context = [[BDRuleContext alloc] initWithModel:myModel];
//
//    [context setValue:@"red" forKey:@"color"];
//    XCTAssertEqualObjects([context valueForKey:@"saturation"], @"75", nil);
//
//    [context setValue:@"green" forKey:@"color"];
//    XCTAssertEqualObjects([context valueForKey:@"saturation"], @"50", nil);
//
//    [context setValue:@"blue" forKey:@"color"];
//    XCTAssertEqualObjects([context valueForKey:@"saturation"], @"66", nil);
//
//    [context release];
//    [myModel release];
//}

- (void)testKeyAssignment
{
    BDRuleModel *testModel;
    BDRule *fooRule;
    BDRuleContext *context;

    testModel = [[BDRuleModel alloc] init];

    fooRule = [BDRule ruleWithLhs:[NSPredicate predicateWithFormat:@"foo == foo"]
                              rhs:[BDKeyAssignment assignmentWithKeyPath:@"bar"
                                                                   value:@"baz"]
                         priority:0];
    
    [testModel addToRules:fooRule];

    context = [[BDRuleContext alloc] initWithModel:testModel];

    [context setValue:@"foo"
                forKey:@"foo"];
    
    [context setValue:@"quux"
                forKey:@"baz"];

    XCTAssertEqualObjects([context valueForKey:@"bar"], @"quux", @"");

}


- (void)testPropertyListAssignment
{
    BDRuleModel *testModel;
    BDRule *fooRule;
    BDRuleContext *context;

    testModel = [[BDRuleModel alloc] init];

    fooRule = [BDRule ruleWithLhs:[NSPredicate predicateWithFormat:@"foo == foo"]
                              rhs:[BDPropertyListAssignment assignmentWithKeyPath:@"bar"
                                                                            value:
                                   @"{ x = x1; y = y1; z = (z1, z2); }"]
                         priority:0];
    
    [testModel addToRules:fooRule];

    context = [[BDRuleContext alloc] initWithModel:testModel];

    [context setValue:@"foo"
                forKey:@"foo"];

    [context setValue:@"quux"
                forKey:@"baz"];

    XCTAssertTrue(([[context valueForKey:@"bar"] isEqualTo:[NSDictionary dictionaryWithObjectsAndKeys:
        @"x1", @"x",
        @"y1", @"y",
        [NSArray arrayWithObjects:@"z1", @"z2", nil], @"z",
        nil]]), @"");

}

- (void)testRuleModelInitWithPropertyList
{
    NSString *plistRep;
    BDRuleModel *myModel;
    BDRuleContext *context;

    plistRep =
        @"{\n"
        @"    version = 1;\n"
        @"    rules = (\n"
        @"        {\n"
        @"            version = 1;\n"
        @"            lhs = \"color = 'red'\";\n"
        @"            rhs = {\n"
        @"                version = 1;\n"
        @"                keyPath = saturation;\n"
        @"                value = \"75\";\n"
        @"            };\n"
        @"            assignmentClass = BDAssignment;\n"
        @"            priority = 0;\n"
        @"        },\n"
        @"        {\n"
        @"            version = 1;\n"
        @"            lhs = \"color = 'green'\";\n"
        @"            rhs = {\n"
        @"                version = 1;\n"
        @"                keyPath = saturation;\n"
        @"                value = \"50\";\n"
        @"            };\n"
        @"            assignmentClass = BDAssignment;\n"
        @"            priority = 0;\n"
        @"        },\n"
        @"        {\n"
        @"            version = 1;\n"
        @"            lhs = \"color = 'blue'\";\n"
        @"            rhs = {\n"
        @"                version = 1;\n"
        @"                keyPath = saturation;\n"
        @"                value = \"66\";\n"
        @"            };\n"
        @"            assignmentClass = BDAssignment;\n"
        @"            priority = 0;\n"
        @"        }\n"
        @"    );\n"
        @"}\n";
    
    myModel = [[BDRuleModel alloc] initWithPropertyList:[plistRep propertyList]
                                                  owner:nil];

    XCTAssertTrue([[myModel rules] count] == 3, @"");
    
    context = [[BDRuleContext alloc] initWithModel:myModel];


    [context setValue:@"red"
                forKey:@"color"];
    XCTAssertEqualObjects([context valueForKey:@"saturation"], @"75", @"");

    [context setValue:@"green"
                forKey:@"color"];
    XCTAssertEqualObjects([context valueForKey:@"saturation"], @"50", @"");

    [context setValue:@"blue"
                forKey:@"color"];
    XCTAssertEqualObjects([context valueForKey:@"saturation"], @"66", @"");

}

- (void)checkPlistRule:(NSMutableDictionary *)plist
                   lhs:(NSString *)lhs
          rhsClassName:(NSString *)rhsClassName
            rhsKeyPath:(NSString *)rhsKeyPath
              rhsValue:(id)rhsValue
              priority:(int)priority
{
    XCTAssertTrue([plist count] == 5, @"");
    XCTAssertTrue([[plist valueForKey:@"version"] intValue] == 1, @"");
    XCTAssertEqualObjects([plist valueForKey:@"lhs"], lhs, @"");
    XCTAssertEqualObjects([plist valueForKey:@"assignmentClass"], rhsClassName, @"");

    XCTAssertTrue([[plist valueForKeyPath:@"rhs.version"] intValue] == 1, @"");
    XCTAssertEqualObjects([plist valueForKeyPath:@"rhs.keyPath"], rhsKeyPath, @"");
    XCTAssertEqualObjects([plist valueForKeyPath:@"rhs.value"], rhsValue, @"");

    XCTAssertTrue([[plist valueForKey:@"priority"] intValue] == priority, @"");
}

- (void)testRuleModelEncodeIntoPropertyList
{
    NSMutableDictionary *plist;

    plist = [NSMutableDictionary dictionaryWithCapacity:0];

    [model encodeIntoPropertyList:plist];

    XCTAssertTrue([plist count] == 2, @"");

    XCTAssertTrue([[plist valueForKey:@"version"] intValue] == 1, @"");
    XCTAssertTrue([[plist valueForKey:@"rules"] count] == 6, @"");

    [self checkPlistRule:[[plist valueForKey:@"rules"] objectAtIndex:0]
                     lhs:@"stock.differenceFromPreviousDayClose > 0"
            rhsClassName:@"BDAssignment"
              rhsKeyPath:@"color"
                rhsValue:@"green"
                priority:0];

    [self checkPlistRule:[[plist valueForKey:@"rules"] objectAtIndex:1]
                     lhs:@"stock.differenceFromPreviousDayClose < 0"
            rhsClassName:@"BDAssignment"
              rhsKeyPath:@"color"
                rhsValue:@"red"
                priority:0];

    [self checkPlistRule:[[plist valueForKey:@"rules"] objectAtIndex:2]
                     lhs:@"TRUEPREDICATE"
            rhsClassName:@"BDAssignment"
              rhsKeyPath:@"color"
                rhsValue:@"black"
                priority:0];

    [self checkPlistRule:[[plist valueForKey:@"rules"] objectAtIndex:3]
                     lhs:@"color == \"red\""
            rhsClassName:@"BDAssignment"
              rhsKeyPath:@"hue"
                rhsValue:[[NSNumber numberWithDouble:0.0] description]
                priority:0];

    [self checkPlistRule:[[plist valueForKey:@"rules"] objectAtIndex:4]
                     lhs:@"color == \"green\""
            rhsClassName:@"BDAssignment"
              rhsKeyPath:@"hue"
                rhsValue:[[NSNumber numberWithDouble:120.0] description]
                priority:0];

    [self checkPlistRule:[[plist valueForKey:@"rules"] objectAtIndex:5]
                     lhs:@"TRUEPREDICATE"
            rhsClassName:@"BDAssignment"
              rhsKeyPath:@"hue"
                rhsValue:[[NSNumber numberWithDouble:240.0] description]
                priority:0];
}

@end
